/* 
 * RESDIR: Search and replace strings in a GEM resource file.
 * Copyright (C) 2002 John Elliott <jce@seasip.demon.co.uk>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ppdgem.h"

#define RT_VRSN 0
#define RT_OB 1
#define RT_TEDINFO 2
#define RT_ICONBLK 3
#define RT_BITBLK 4
#define RT_FREESTR 5
#define RT_STRING 6
#define RT_IMAGEDATA 7
#define RT_FREEIMG 8
#define RT_TRINDEX 9

#define NUM_RTYPES 10

#define R_NOBS 10
#define R_NTREE 11
#define R_NTED 12
#define R_NICON 13
#define R_NBITBLK 14
#define R_NSTRING 15
#define R_IMAGES 16
#define NUM_RN 7

#define RS_SIZE 17                              /* NUM_RTYPES + NUM_RN  */

#define HDR_LENGTH (RS_SIZE + 1) * 2            /* in bytes     */

#ifndef SEEK_SET
#define SEEK_SET 0
#endif

MLOCAL WORD rs_hdr[RS_SIZE + 1];

MLOCAL void *tedinfos, *images;
MLOCAL OBJECT *objects;
MLOCAL ICONBLK *iconblks;
MLOCAL BITBLK *bitblks;
MLOCAL LONG *frimages;
MLOCAL LONG *treeindex;
MLOCAL LONG *freestrs;
MLOCAL char *strings;
MLOCAL UWORD strs_len;

MLOCAL void free_sections()
{
	if (objects)   free(objects);   objects   = NULL;
	if (tedinfos)  free(tedinfos);  tedinfos  = NULL;
	if (iconblks)  free(iconblks);  iconblks  = NULL;
	if (bitblks)   free(bitblks);   bitblks   = NULL;
	if (images)    free(images);    images    = NULL;
	if (frimages)  free(frimages);  frimages  = NULL;
	if (treeindex) free(treeindex); treeindex = NULL;
	if (freestrs)  free(freestrs);  freestrs  = NULL;
	if (strings)   free(strings);   strings   = NULL;
}

MLOCAL WORD load_section(FILE *fp, UWORD sect, UWORD len, void **data)
{
	sect  = rs_hdr[sect];

	*data = malloc(len);
	if (*data == NULL) { fprintf(stderr, "Out of memory\n"); return -3; }
	fseek(fp, sect, SEEK_SET);
	if (fread(*data, 1, len, fp) < len) 
	{ 
		fprintf(stderr, "Unexpected EOF\n"); 
		free(*data); 
		*data = NULL; 
		return -4; 
	}
	return 0;
}



MLOCAL WORD save_section(FILE *fp, UWORD sect, UWORD len, void *data)
{
	sect  = rs_hdr[sect];

	fseek(fp, sect, SEEK_SET);
	if (fwrite(data, 1, len, fp) < len) 
	{ 
		fprintf(stderr, "Failed to write\n"); 
		return -4; 
	}
	return 0;
}




WORD load_rsc(char *filename)
{
	FILE *fp = fopen(filename, "rb");
	if (!fp) 
	{ 	
		perror(filename); 
		return -1; 
	}

	if (fread(rs_hdr, 1, HDR_LENGTH, fp) < HDR_LENGTH)
	{
		fclose(fp);
		fprintf(stderr, "Could not load RSC header from %s\n", filename);
		return -2;
	}
	// Load sections
	if (load_section(fp, RT_OB,        rs_hdr[R_NOBS]    * sizeof(OBJECT),         (void **)&objects)
	||  load_section(fp, RT_TEDINFO,   rs_hdr[R_NTED]    * sizeof(TEDINFO),        &tedinfos)
	||  load_section(fp, RT_ICONBLK,   rs_hdr[R_NICON]   * sizeof(ICONBLK),        (void **)&iconblks)
	||  load_section(fp, RT_BITBLK,    rs_hdr[R_NBITBLK] * sizeof(BITBLK),         (void **)&bitblks)
	||  load_section(fp, RT_FREESTR,   rs_hdr[R_NSTRING] * 4,                      (void **)&freestrs)
	||  load_section(fp, RT_STRING,    rs_hdr[RT_IMAGEDATA] - rs_hdr[RT_STRING],   (void **)&strings)
	||  load_section(fp, RT_IMAGEDATA, rs_hdr[RT_FREESTR]   - rs_hdr[RT_IMAGEDATA],&images)
	||  load_section(fp, RT_FREEIMG,   rs_hdr[R_IMAGES]  * 4,                      (void **)&frimages)
	||  load_section(fp, RT_TRINDEX,   rs_hdr[R_NTREE]   * 4,                      (void **)&treeindex))
	{
		free_sections();
		return -3;
	}
	strs_len = rs_hdr[RT_IMAGEDATA] - rs_hdr[RT_STRING];
	fclose(fp);
	return 0;
}


// Then program simply generates a new string section based on the existing 
// one, substituting new GEMAPPS and GEMSYS if appropriate.

char *get_string(int idx)
{
	WORD w = freestrs[idx];

	return strings + w - HDR_LENGTH;
}


int string_count(void)
{
	return rs_hdr[R_NSTRING];
}


void put_string(int idx, char *s)
{
	char *orig = get_string(idx);
	char *buf;
	unsigned long orig_len = strlen(orig) + 1;
	unsigned long new_len  = strlen(s)    + 1;
	unsigned long buf_len = strs_len - orig_len + new_len;
	unsigned long offset = freestrs[idx] - HDR_LENGTH;
	unsigned long n;

	buf = malloc(buf_len);
	if (!buf) return;
	memcpy(buf, strings, offset);
	strcpy(buf + offset, s);
	memcpy(buf + offset + new_len, strings + offset + orig_len, strs_len - offset - orig_len);
	
	for (n = idx + 1; n < rs_hdr[R_NSTRING]; n++)
	{
		freestrs[n] += (new_len - orig_len);
	}
	strs_len = buf_len;
	free(strings);
	strings = buf;
/* The strings section has been resized. Change other pointers accordingly */
	for (n = RT_OB; n < RT_STRING; n++)
	{
		rs_hdr[n] += (new_len - orig_len);
	}
	rs_hdr[RT_IMAGEDATA] += (new_len - orig_len);
	rs_hdr[RT_FREEIMG]   += (new_len - orig_len);
	rs_hdr[RT_TRINDEX]   += (new_len - orig_len);

	for (n = 0; n < rs_hdr[R_NTREE]; n++)
	{
		treeindex[n] += (new_len - orig_len);
	}
	for (n = 0; n < rs_hdr[R_IMAGES]; n++)
	{
		frimages[n] += (new_len - orig_len);
	}
	for (n = 0; n < rs_hdr[R_NBITBLK]; n++)
	{
		bitblks[n].bi_pdata += (new_len - orig_len);
	}
	for (n = 0; n < rs_hdr[R_NICON]; n++)
	{
		iconblks[n].ib_pdata += (new_len - orig_len);
		iconblks[n].ib_pmask += (new_len - orig_len);
	}
	for (n = 0; n < rs_hdr[R_NOBS]; n++)
	{
		switch(objects[n].ob_type & 0xFF)
		{
			LONG l;

			case G_TEXT: case G_BOXTEXT: case G_IMAGE: 
			case G_STRING: case G_BUTTON:
			case G_FTEXT: case G_FBOXTEXT:
			case G_ICON: case G_CLRICN:
			l = (LONG)(objects[n].ob_spec);
			if (l > (offset + HDR_LENGTH))
				l += (new_len - orig_len);
			objects[n].ob_spec = (void *)l;
		}
	}
	rs_hdr[RS_SIZE] += (new_len - orig_len);
}

WORD save_rsc(char *filename)
{
	FILE *fp = fopen(filename, "wb");
	if (!fp) 
	{ 	
		perror(filename); 
		return -1; 
	}

	if (fwrite(rs_hdr, 1, HDR_LENGTH, fp) < HDR_LENGTH)
	{
		fclose(fp);
		fprintf(stderr, "Could not save RSC header to %s\n", filename);
		return -2;
	}
	// Load sections
	if (save_section(fp, RT_OB,        rs_hdr[R_NOBS]    * sizeof(OBJECT),         objects)
	||  save_section(fp, RT_TEDINFO,   rs_hdr[R_NTED]    * sizeof(TEDINFO),        tedinfos)
	||  save_section(fp, RT_ICONBLK,   rs_hdr[R_NICON]   * sizeof(ICONBLK),        iconblks)
	||  save_section(fp, RT_BITBLK,    rs_hdr[R_NBITBLK] * sizeof(BITBLK),         bitblks)
	||  save_section(fp, RT_FREESTR,   rs_hdr[R_NSTRING] * 4,                      freestrs)
	||  save_section(fp, RT_STRING,    rs_hdr[RT_IMAGEDATA] - rs_hdr[RT_STRING],   strings)
	||  save_section(fp, RT_IMAGEDATA, rs_hdr[RT_FREESTR]   - rs_hdr[RT_IMAGEDATA],images)
	||  save_section(fp, RT_FREEIMG,   rs_hdr[R_IMAGES]  * 4,                      frimages)
	||  save_section(fp, RT_TRINDEX,   rs_hdr[R_NTREE]   * 4,                      treeindex))
	{
		return -3;
	}
	fclose(fp);
	return 0;
}


